var ShortcutsCtrl = function(w, parent, LegacyExt, Ext, dashEvents) {

    /**
     * Type WSDashboard
     * | acl : String
     * | allmayview : String
     * | description : String
     * | image : String
     * | last_revision : String
     * | name : String
     * | uuid : String
     * | username : String
     */

    /**
     * Type Meta
     * | type_id : String
     * | element_id : String
     */

    /**
     * Sources
     * Sources are the metadata that will represent information for the function composition to manage the shortcuts events
     */
    var sources =
        { url: "/opmon/seagull/www/index.php/opdashboard/action/getDashboardsTooltip"
        , idsToShow: [ 4, 6 ]
        , folderPreventActions: "OpMon"
        , id: "188"
        };

    var State = function(bool) {
        return (
            { alreadyExecuted: function() { return bool; }
            , next: function(x) {
                return (typeof x === "boolean")
                    ? (bool = x, this)
                    : (bool = !bool, this);
              }
            , fold: function(f, g) { return bool ? g () : f (); }
            }
        );
    }

    var initState =
        State (false);

    function bootEvents() {
        return (
            { afterrender: initTooltip.bind(null, sources)
            , rightclick: rightClick.bind(null, sources)
            , beforedrop: beforedrop.bind(null, sources)
            , config: config.bind(null, sources, initState)
            }
        );
    }

    /**
     * Boot
     * Boot fires a function at the time the window and shortcuts were loaded
     */
    function boot(fs) {
        return (parent.document.readyState === "complete")
            ? callParentLoader(fs)
            : parent.addEventListener("load", callParentLoader.bind(null, fs));
    }


    /**
     * callParentLoader
     * Will fire the function that loads the shortcuts
     */
    function callParentLoader(fs) {
        parent.loadShortcut('dashboards', null, fs);
		parent.removeContextSentitivePanel();
    }


    function initTooltip(srcs, tree) {
        var tooltips =
            tree.getNodes()
                .map(createTooltip (srcs));
    }


    function rightClick(srcs, node) {
        return w.op.Maybe.of (node)
            .chain (getFolderRootName)
            .fold
                ( function() { return true; }
                , function(val) { return (val !== srcs.folderPreventActions) }
                );
    }

    function beforedrop(srcs, tree, data, target, op) {
        return w.op.Maybe.of (
            w.op.Functions.curry(function(x, y) { return [x, y]; })
        )
            .ap (helpDropGetFolderRootName (data))
            .ap (getFolderRootName (target))
            .map (w.op.List.all (w.op.Functions.diff (srcs.folderPreventActions)))
            .map (w.op.Functions.and (targetIsntRoot (target, op)))
            .fold
                ( w.op.Functions.always (false)
                , w.op.Functions.identity
                );
    }

    function config(srcs, state, tree) {
        return (state.alreadyExecuted ())
            ? (collapseOnly (srcs.folderPreventActions, tree), state)
            : (collapseAll (srcs.id, tree), state.next (), state);
    }

    function collapseAll(id, tree) {
        var predicate = function(x) { return x.data.id === id; };
        tree.collapseAll(forEachNodeApplyIf.bind(null, predicate, "expand", tree));
    }

    function collapseOnly(text, tree) {
        var predicate = function(x) { return x.data.text === text; };
        forEachNodeApplyIf (predicate, "collapse", tree);
    }

    function forEachNodeApplyIf (logicPredicate, f, tree)  {
        tree.getRootNode().childNodes.forEach(function(x) {
            if (logicPredicate (x) && x[f]) {
                x[f]();
            }
        });
    }

    function helpDropGetFolderRootName(node) {
        return w.op.Maybe.of (node)
            .chain (w.op.Object.pluck ("records"))
            .chain (w.op.List.head)
            .chain (getFolderRootName)
    }


    var getFolderRootName =
        w.op.Functions.pipe
            ([ findFolderRoot
             , w.op.Object.deepPluck ([ "data", "text" ])
            ])


    function findFolderRoot(n) {
        return (n.parentNode.data.root)
            ? n
            : findFolderRoot (n.parentNode);
    }


    function targetIsntRoot(target, op) {
        return (op === "before" && target.parentNode.data.root) === false;
    }


    // createTooltip :: Source -> State -> Html Node -> Task (LegacyExt Tooltip)
    var createTooltip = w.op.Functions.curry(function(srcs, node) {
        return extractMetas(srcs.idsToShow, node)
                .fold
                    ( basicTooltip (node)
                    , tooltip (srcs.url, node)
                    );
    });

    // String -> Html Node -> Meta -> LegacyExt Tooltip
    var tooltip = w.op.Functions.curry(function(url, node, params) {
        if (LegacyExt.get(node).hasTooltip) return false;
        LegacyExt.get(node).hasTooltip = true
        var ttip = new LegacyExt.tip.ToolTip({
            target: node,
            mouseOffset: [0, -5],
            width: 400,
            showDelay: 1000,
            dismissDelay: 0,
            hideDelay: 0,
            autoHide: false,
            bodyBorder: false,
            frame: false,
            baseCls: "object-tooltip",
            loader: {
                url: url,
                method: "POST",
                params: params,
                text: "Loading...",
                nocache: true
            },
            listeners: {
                beforeshow: function(tip) {
                    tip.loader.load();
                }
            }
        });

        var timer = createTimer (500, ttip.hide.bind (ttip));

        ttip.on("show", function() {
            var el =
                removeListeners ([ ["mouseover", timer.cancel], ["mouseout", timer.next] ], ttip.getEl().dom);

            var nodeEl =
                removeListeners ([ ["mouseover", timer.cancel], ["mouseout", timer.next] ], node);

            el.addEventListener ( "mouseover", timer.cancel );
            nodeEl.addEventListener ( "mouseover", timer.cancel );
            el.addEventListener ( "mouseout", timer.next );
            nodeEl.addEventListener ( "mouseout", timer.next );
        });

        return ttip;
    });

    var basicTooltip = w.op.Functions.curry(function(node, text) {
        if (LegacyExt.get(node).hasTooltip) return false;
        LegacyExt.get(node).hasTooltip = true
        var ttip = new LegacyExt.tip.ToolTip({
            target: node,
			html: node.textContent,
            mouseOffset: [0, -5],
            showDelay: 1000,
            dismissDelay: 0,
            hideDelay: 0,
            autoHide: true,
            bodyBorder: false,
            frame: false,
			baseCls: "object-tooltip",
        });

        return ttip;
    });


    // createTimer :: Number -> Function -> {next, cancel}
    var createTimer = w.op.Functions.curry(function(n, f) {
        var timer;
        return (
            { next: function timerNext() {
                    var args = Array.prototype.slice.call(arguments)
                    timer = setTimeout(function() {
                        f.apply(null, args)
                    }, n);
                }
            , cancel: function cancelTimer() {
                    clearTimeout (timer);
                }
            }
        );
    });

    // removeListeners :: [[String, Function]] -> Html Nodo -> Html Nodo
    var removeListeners = w.op.Functions.curry(function(xs, nodo) {
        return xs.reduce (function(n, x) {
            return (n.removeEventListener (fst(x), scd(x)), n);
        }, nodo);
    });


    // fst :: [a] -> a
    function fst(xs) { return xs[0]; }
    // scd :: [a] -> a
    function scd(xs) { return xs[1]; }


    // extractMetas :: [Number] -> Node html -> Either String Meta
    var extractMetas = w.op.Functions.curry(function(ids, node) {
        return getHrefData (node)
                    .map (getMetas)
                    .chain (verifyIds (ids))
                    .map (applyName (node))
                    .map (applyMode ("OPVIEW"));
    });


    // getHrefData :: Node html -> Either String String
    function getHrefData(node) {
        return w.op.Either.of (node)
            .map (function(x) {
                return x.querySelector("a")
                    .getAttribute("href")
                    .match(/openDashInFlash\((.*)\)/);
            })
            .chain (function(xs) {
                return (xs)
                    ? w.op.Either.of(xs [1])
                    : w.op.Either.Left ("has no meta");
            });
    }


    // getMetas :: String -> Meta
    var getMetas =
        w.op.Functions.pipe
            ([ w.op.String.trim
             , w.op.String.split(',')
             , function(xs) { return xs.map (w.op.String.trim); }
             , reduceMetas
            ]);


    // verifyIds :: [Number] -> Meta -> Either String Meta
    var verifyIds = w.op.Functions.curry(function(ids, meta) {
        return (ids.some(function(id) { return (id === meta.type_id); }))
                    ? w.op.Either.Right (meta)
                    : w.op.Either.Left ("Cannot show tooltip to this type");
    });


    var apply = w.op.Functions.curry(function(key, o, x) {
        return LegacyExt.apply (o, w.op.Object.assign (key, x, o));
    });


    // applyName :: Node html -> Meta -> Either String Meta
    var applyName = w.op.Functions.curry(function(node, o) {
        return apply ("name", o, node.querySelector("a").innerHTML);
    });


    // applyMode :: Node html -> String -> Either String Meta
    var applyMode = w.op.Functions.curry(function(x, o) {
        return apply ("mod", o, x);
    });


    // reduceMetas :: [String] -> Meta
    function reduceMetas(xs) {
        return xs.reduce(function(acc, x, i) {
            return (i === 0)
                ? w.op.Object.assign ("type_id", Number(x)) (acc)
                : w.op.Object.assign ("element_id", w.op.List.removeFirstLast(x)) (acc);
        }, {});
    }

    function commandToOpenDashboard(type, id, isEdit) {
        var iframe = Array.prototype.slice.call(document.getElementsByTagName("iframe"));
        var dashContainer = Ext.ComponentQuery.query("#dashPanel");
        var perform = parent.zip(function(iframe, container) {
            container.setActiveItem(1);
            var timer = setInterval(function() {
                if (iframe.dashLoaded) {
                    clearInterval(timer);
                    return openDash(iframe);
                }
            }, 500);
        });
        
        function openDash(iframe) {
            var params = id === "newDashboard" ? [] : [type, id];
            var context = iframe.contentWindow.document.mainEmbed;
            context.openDashboard.apply(context, params);
            if (isEdit) {
                setTimeout(context.openEditor.bind(context), 500);
            }
        }
        
        perform(iframe, dashContainer);
    }

    var dashTimer = createTimer (500, commandToOpenDashboard);

    parent.openDashInFlash = function(timer, type, id) {
        timer.cancel();
        timer.next(type, id);
    }.bind(null, dashTimer);
    
    dashEvents.subscribe({
        id: ["dashboard_saved", "dashboard_removed"],
        action: boot.bind(null, bootEvents())
    });
    
    dashEvents.subscribe({
        id: "dashboard_complete",
        action: function(_, frame) {
            frame.dashLoaded = true;
        }
    });
    
    return {
        boot: boot,
        bootEvents: bootEvents,
        commandToOpenDashboard: commandToOpenDashboard,
    };

};
