(function(w) {

    w.op = w.op || {};
    
}) (window);
(function(w, Ext) {

    /**
     * Functions Utils
     */

    function curry(fx) {
        var arity = fx.length;
        return function f1() {
            var args = Array.prototype.slice.call(arguments, 0);
            return (args.length >= arity)
            ? fx.apply(null, args)
            : function() {
                var args2 = Array.prototype.slice.call(arguments, 0);
                return f1.apply(null, args.concat(args2));
            };
        };
    }

    // compose :: [(a -> b)] -> a -> b
    function compose(fs) {
        return function(init) {
            return fs.reduceRight(function(res, f) { return f(res); }, init);
        }
    }

    // pipe :: [(a -> b)] -> a -> b
    function pipe(fs) {
        return function(init) {
            return fs.reduce(function(res, f) { return f(res); }, init);
        }
    }

    // always :: a -> _ -> a
    var always = curry(function(x, _) { return x; })
    
    // K :: a -> _ -> a
    var K = function(x) {
        return function() {
            return x;
        }
    }

    // identity :: a -> a
    function identity(x) { return x; }

    // log :: String -> a -> a
    var log = curry(function(msg, x) {
        return (console.warn("## LOG ## " + msg + ": ", x), x);
    });

    // equals :: a -> b -> Bool
    var equals = curry(function(x, y) { return (x === y); });

    // diff :: a -> b -> Bool
    var diff = curry(function(x, y) { return !equals (x, y); });

    // and :: a -> b -> Bool
    var and = curry(function(x, y) { return x && y; });

    // or :: a -> b -> Bool
    var or = curry(function(x, y) { return x || y; });

    w.op.Functions =
        { compose: compose
        , pipe: pipe
        , curry: curry
        , always: always
        , identity: identity
        , log: log
        , equals: equals
        , diff: diff
        , and: and
        , or: or
        , K: K
        }


    /**
     * Object Utils
     */

    // assign :: String -> a -> Object a -> Object a
    var assign = curry(function(key, value, o) {
        var fresh = Ext.clone(o);
        fresh[key] = value;
        return fresh;
    });

    // pluck :: String -> Object a -> Maybe a
    var pluck = curry(function(key, o) {
        return w.op.Maybe.of (o [key]);
    });

    // deepPluck :: [String] -> Object a -> a
    var deepPluck = curry(function(ks, object) {
        return ks.reduce(function(ma, k) {
            return ma.chain (pluck (k));
        }, w.op.Maybe.of (object));
    });

    w.op.Object =
        { assign: assign
        , pluck: pluck
        , deepPluck: deepPluck
        }


    /**
     * List Utils
     */

    // removeFirstLast :: List a -> List a
    function removeFirstLast(str) {
        return str.slice(1).slice(0,-1);
    }

    // last :: List a -> a
    function last(xs) {
        return xs.slice(-1);
    }

    // head :: List a -> Maybe a
    function head(xs) {
        return w.op.Maybe.of (xs [0]);
    }

    // all :: (a -> Bool) -> [a] -> Bool
    var all = curry(function(f, xs) {
        return xs.reduce(function(acc, x) {
            return acc && f (x);
        }, true);
    });

    // any :: (a -> Bool) -> [a] -> Bool
    var any = curry(function(f, xs) {
        return xs.reduce(function(acc, x) {
            return acc || f (x);
        }, false);
    });
    
    var zip = function(f) {
        return function() {
            var xxs = Array.prototype.slice.call(arguments);
            return lowerArray(xxs)
                .map(function(_, i) {
                    var args = xxs.map(function(xs) { return xs[i]; });
                    return f.apply(null, args);
                });
        };
    };

    var lowerArray = function(xxs) {
        return xxs.reduce(function(lower, next) {
            return next.length < lower.length ? next : lower;
        });
    };


    w.op.List =
        { removeFirstLast: removeFirstLast
        , last: last
        , head: head
        , all: all
        , any: any
        , zip: zip
        , lowerArray: lowerArray
        }
    

    /**
     * String Util
     */

    // trim :: String -> String
    function trim(str) {
        return str.trim();
    }

    // split :: Regex -> String -> [String]
    var split = curry(function(reg, str) {
        return str.split (reg);
    });

    w.op.String =
        { trim: trim
        , split: split
        }

    var getTwoHours = function() {
        var d = new Date();
        d.setHours(d.getHours() + 2);
        return d;
    }
    
    var formatDateToDb = function(d) {
        return Ext.Date.format(new Date(d), 'Y-m-d')
    };
    
    var formatTimeToDb = function(d) {
        return Ext.Date.format(new Date(d), 'h:i:s')
    };

    var formatDateEN = function(d) {
        return Ext.Date.format(new Date(d), 'm/d/Y')
    };
    
    w.op.Date =
        { getTwoHours: getTwoHours,
          formatDateToDb: formatDateToDb,
          formatTimeToDb: formatTimeToDb,
          formatDateEN: formatDateEN
        }
    
}) (window, (window.Ext || parent.Ext));

(function(w) {

    /**
     * Type Either a | b
     * | Left b
     * | Right a
     */

    function handleErr(f, g, x) {
        try {
            return g (x);
        } catch (e) {
            //console.warn("OP LOG INFO", e);
            return f (e);
        }
    }

    function exec(f, x) {
        return handleErr(Left, function(y) { return Right (f (y)); }, x);
    }

    var Left = function(x) {
        return (
            { x: x
            , isRight: false
            , map: function(f) { return Left (x); }
            , chain: function(f) { return Left (x); }
            , ap: function(other) { return Left (x); }
            , fold: function(f, g) { return f (x); }
            }
        );
    };

    var Right = function(x) {
        return (
            { x: x
            , isRight: true
            , map: function(f) { return exec (f,x); }
            , chain: function(f) { return exec(f,x).fold(Left, function(y) { return y; }); }
            , ap: function(other) { return other.map(x); }
            , fold: function(f, g) { return handleErr(f, g, x); }
            }
        );
    };

    // of :: a -> Either a
    var of = Right;

    // eitherToTask :: Either a -> Task a
    function eitherToTask(either) {
        return either.fold
            ( function(e) { return Task.rejected(e); }
            , function(x) { return Task.of(x); }
            );
    }

    w.op.Either =
        { Left: Left
        , Right: Right
        , of: of
        , eitherToTask: eitherToTask
        , execOf: exec
        };
    
}) (window);



(function(w) {

    /**
     * Type Maybe a
     * | Nothing
     * | Just a
     */

    var Nothing =
        { x: "Nothing"
        , isNothing: true
        , map: function(f) { return Nothing; }
        , chain: function(f) { return Nothing; }
        , ap: function(other) { return Nothing; }
        , fold: function(f, g) { return f (); }
        };

    var Just = function(x) {
        return (
            { x: x
            , isNothing: false
            , map: function(f) { return validate(x) ? Just (f (x)) : Nothing; }
            , chain: function(f) { return validate(x) ? f (x) : Nothing; }
            , ap: function(other) { return other.map(x); }
            , fold: function(f, g) { return validate (x) ? g (x) : f(); }
            }
        );
    };

    // validate :: a -> Bool
    function validate(x) {
        return (x !== null) && (x !== undefined);
    }

    // of :: a -> Maybe a
    function of(x) {
        return validate (x)
            ? Just (x)
            : Nothing
    }

    // withDefault :: b -> Maybe a -> a | b
    var withDefault = w.op.Functions.curry(function(x, ma) {
        return ma.fold(w.op.Functions.always(x), w.op.Functions.identity);
    });

    // maybeToTask :: Maybe a -> Task a
    function maybeToTask(ma) {
        return ma.fold
            ( function(e) { return Task.rejected(e); }
            , function(x) { return Task.of(x); }
            );
    }

    w.op.Maybe =
        { Nothing: Nothing
        , Just: Just
        , of: of
        , withDefault: withDefault
        , maybeToTask: maybeToTask
        };
    
}) (window);